/**
 * ECE 4760: Laboratory 5
 * Smart Station
 * File: main.c
 * Target PIC:PIC32MX250F128B
 * 
 * Authors: Dhruv Gaba, Shaurya Luthra, and Tim Braren
 * 
 * This firmware turns the PIC32 development board into a smart station
 * 
 **/

// Graphics libraries
#include "config.h"
#include "tft_master.h"
#include "tft_gfx.h"
#include "i2c_rw.h"
#include "gui.h"
#include <time.h>
#include "neopixel.h"

// Needed for rand function
#include <stdlib.h>

// PIC32 Peripheral Library
#include <plib.h>

// Config.h sets clock to 40 MHz
#include "pt_cornell_1_2.h"
#include <math.h>

// Define for system frequency
#define	SYS_FREQ 40000000

// Defines for de-bounce FSM
#define NO_PUSH 0
#define MAYBE_PUSH 1
#define PUSHED 2
#define MAYBE_NO_PUSH 3

// Sensor Values
float imu_values[3];
short luminosity_values[2];
float temp_humidity[2];

#define buf_size 10

int buffer_pow[buf_size];
int buf_idx = 0;
int buf_power = 0;
int screen_fsm = MAIN_SCREEN;
int changed=0;
unsigned char slave_address;

int react_music = 0;

// === Protothread Structures ==================================================
static struct pt pt_input, pt_timer, pt_DMA_output, write_neo, measure_thread;

// Thread responsible for outputting data on TFT display
static PT_THREAD(protothread_timer(struct pt *pt)) 
{
  PT_BEGIN(pt);
  static int timer_start;
  while (1) 
  {
    timer_start = PT_GET_TIME();
    switch (screen_fsm) 
    {
      case MAIN_SCREEN:
        main_screen();
        PT_YIELD_UNTIL(pt, PT_GET_TIME() - timer_start >= 500);
        break;

      case SETTINGS_SCREEN:
        settings_screen();
        PT_YIELD_UNTIL(pt, PT_GET_TIME() - timer_start >= 500);
        break;

      case MUSIC_SCREEN:
        music_screen();
        PT_YIELD_UNTIL(pt, PT_GET_TIME() - timer_start >= 500);
        break;

      case GRAPH_SCREEN:
        graph_screen();
        PT_YIELD_UNTIL(pt, PT_GET_TIME() - timer_start >= 200);
        break;
    }
  }
  PT_END(pt);
}

// === Input Thread ============================================================
// connections:
// A0 -- row 1 -- thru 300 ohm resistor -- avoid short when two buttons pushed
// A1 -- row 2 -- thru 300 ohm resistor
// A2 -- row 3 -- thru 300 ohm resistor
// A3 -- row 4 -- thru 300 ohm resistor
// B7 -- col 1 -- 10k pulldown resistor -- avoid open circuit input when no 
//                                      -- button pushed
// B8 -- col 2 -- 10k pulldown resistor
// B9 -- col 3 -- 10k pulldown resistor

static PT_THREAD(protothread_input(struct pt *pt)) 
{
  PT_BEGIN(pt);
  
  // Order of array is 0 through 9 followed by * == 10 and # == 11
  // Absence of a press is encoded with "-1"
  // Table is decoded to natural digit order (except for * and #)
  static int keytable[12] ={
    0x1100,  // 0
    0x28,    // 1
    0x88,    // 2
    0x408,   // 3
    0x50,    // 4
    0x110,   // 5
    0x810,   // 6
    0x140,   // 7
    0x440,   // 8
    0x2040,  // 9
    0x500,   // *
    0x8100}; // #

  static int timer_start, keypad, pattern, i;

  // Variable for maintaining de-bouncing state machine state
  static int state = 0;

  // Encodes whether or not one of the tactile buttons has been pressed
  static int press = 0;

  // Initialize the analog pulse keypad pins RB3
  mPORTBSetPinsDigitalOut(BIT_3);

  // Initialize the keypad input pins RB5 RB7 RB10
  mPORTBSetPinsDigitalIn(BIT_5 | BIT_7 | BIT_10);

  // Initialize MUX control signal pins RB6 RB12
  mPORTBSetPinsDigitalOut(BIT_4 | BIT_15);

  while (1) 
  {
    // Get time that thread began so that length can be limited to 1/20th second
    timer_start = PT_GET_TIME();

    // Read out each row sequentially
    mPORTBClearBits(BIT_4 | BIT_15);
    mPORTBSetBits(BIT_3);

    // Yield for 30 msec
    PT_YIELD_TIME_msec(20);

    while (1) 
    {
      keypad = mPORTBReadBits(BIT_5 | BIT_7 | BIT_10);
      if (keypad != 0) 
      {
        keypad |= BIT_3;
        break;
      }

      mPORTBSetBits(BIT_15);
      PT_YIELD_TIME_msec(20);

      keypad = mPORTBReadBits(BIT_5 | BIT_7 | BIT_10);
      if (keypad != 0) 
      {
        keypad |= BIT_3;
        keypad <<= 1;
        break;
      }

      mPORTBClearBits(BIT_15);
      mPORTBSetBits(BIT_4);
      PT_YIELD_TIME_msec(20);

      keypad = mPORTBReadBits(BIT_5 | BIT_7 | BIT_10);
      if (keypad != 0) 
      {
        keypad |= BIT_3;
        keypad <<= 3;
        break;
      }

      mPORTBSetBits(BIT_15);
      PT_YIELD_TIME_msec(20);

      keypad = mPORTBReadBits(BIT_5 | BIT_7 | BIT_10);
      if (keypad != 0) 
      {
        keypad |= BIT_3;
        keypad <<= 5;
        break;
      }
      break;
    }

    mPORTBClearBits(BIT_3);
    mPORTBClearBits(BIT_4 | BIT_15);

    // Search for key code in key table
    if (keypad > 0) 
    {
      // Some combination of buttons has been pushed
      for (i = 0; i < 12; i++) 
      {
        if (keytable[i] == keypad) 
        {
          break;
        }
      }
    } 
    else 
    {
      // No buttons have been pushed
      i = -1;
    }

    // If any number buttons or the test frequency button have been pressed then
    // set a global press flag.
    if (i >= 0 && i < 12) 
    {
      press = 1;
    }
    else 
    {
      press = 0;
    }

    // State variable possibilities:
    // 0 is no push
    // 1 is maybe push
    // 2 is pushed
    // 3 is maybe no push
    switch (state) 
    {
      case NO_PUSH:
        state = (press ? MAYBE_PUSH : NO_PUSH);
        break;
      case MAYBE_PUSH:
        state = (press ? PUSHED : NO_PUSH);
        if (state == PUSHED) 
        {
          switch (screen_fsm) 
          {
            case MAIN_SCREEN:
              switch (i) 
              {
                case SETTINGS_SCREEN:
                  draw = STATIC;
                  screen_fsm = SETTINGS_SCREEN;
                  break;

                case MUSIC_SCREEN:
                  draw = STATIC;
                  react_music = 1;
                  screen_fsm = MUSIC_SCREEN;
                  break;

                case GRAPH_SCREEN:
                  draw = STATIC;
                  screen_fsm = GRAPH_SCREEN;
                  break;
              }
              break;

            case SETTINGS_SCREEN:
              switch(param)
              {
                case SENSOR:
                  if (i == 10)
                  {
                    draw = STATIC;
                    screen_fsm = MAIN_SCREEN;
                    colorWipeF(0,0,255); 
                    display();
                  }
                  else if (i == 11)
                  {
                    tft_fillCircle(311, (param + 1)*y_offset + 6, 8, ILI9340_BLACK); 
                    param = (param + 1) % 4;
                    tft_fillCircle(311, (param + 1)*y_offset + 6, 8, ILI9340_GREEN);                    
                  }
                  else if (i == 0)
                  {
                    tft_fillRoundRect(88, y_offset-5, 200, 25, 1, ILI9340_BLACK);
                    selected_sensor = (selected_sensor + 1) % 7;
                    tft_setCursor(96, y_offset);
                    tft_setTextColor(ILI9340_WHITE);
                    sprintf(buffer,"%s", sensors[selected_sensor].name);
                    tft_writeString(buffer);
                    
                    tft_fillRoundRect(94, 2*y_offset-3, 19, 19, 1, sensors[selected_sensor].color);
                    
                    tft_fillRoundRect(88, 3*y_offset-5, 200, 25, 1, ILI9340_BLACK);
                    tft_setCursor(96, 3*y_offset);
                    tft_setTextColor(ILI9340_WHITE);
                    sprintf(buffer,"%.2f", sensors[selected_sensor].trig_high);
                    tft_writeString(buffer);
                    
                    tft_fillRoundRect(88, 4*y_offset-5, 200, 25, 1, ILI9340_BLACK);
                    tft_setCursor(96, 4*y_offset);
                    tft_setTextColor(ILI9340_WHITE);
                    sprintf(buffer,"%.2f", sensors[selected_sensor].trig_low);
                    tft_writeString(buffer);
                  }
                  break;

                case COLOR:
                  if (i == 10)
                  {
                    draw = STATIC;
                    screen_fsm = MAIN_SCREEN;
                  }
                  else if (i == 11)
                  {
                    tft_fillCircle(311, (param + 1)*y_offset + 6, 8, ILI9340_BLACK); 
                    param = (param + 1) % 4;
                    tft_fillCircle(311, (param + 1)*y_offset + 6, 8, ILI9340_GREEN);
                    entered=sensors[selected_sensor].trig_high;
                    changed=0;
                  }
                  else if (i == 0)
                  {
                    color_idx = (color_idx + 1) % 7;
                    sensors[selected_sensor].color = colors[color_idx];
                    tft_fillRoundRect(94, 2*y_offset-3, 19, 19, 1, colors[color_idx]);
                  }
                  break;

                case TRIG_HIGH:
                  if (i == 10)
                  {
                    draw = STATIC;
                    screen_fsm = MAIN_SCREEN;
                  }
                  else if (i == 11)
                  {
                    if (entered >= sensors[selected_sensor].min && 
                        entered <= sensors[selected_sensor].max &&
                        entered >= sensors[selected_sensor].trig_low)
                    {
                      sensors[selected_sensor].trig_high = entered;
                      tft_fillCircle(311, (param + 1)*y_offset + 6, 8, ILI9340_BLACK); 
                      param = (param + 1) % 4;
                      tft_fillCircle(311, (param + 1)*y_offset + 6, 8, ILI9340_GREEN);
                      entered=sensors[selected_sensor].trig_low;
                      changed=0;

                    }
                    else
                    {
                      tft_fillCircle(311, (param + 1)*y_offset + 6, 8, ILI9340_RED);
                      entered=0;
                    }
                  }
                  else if (i >= 0 && i <= 9)
                  {
                    if(!changed){
                        entered=0;
                        changed=1;
                    } 
                    // Display number that has been pressed on the keypad
                    entered = entered*10 + i;
                    tft_fillRoundRect(88, 3*y_offset-5, 200, 25, 1, ILI9340_BLACK);
                    tft_setCursor(96, 3*y_offset);
                    tft_setTextColor(ILI9340_WHITE);
                    sprintf(buffer,"%.2f", entered);
                    tft_writeString(buffer);
                  } 
                  break;

                case TRIG_LOW:
                  if (i == 10)
                  {
                    draw = STATIC;
                    screen_fsm = MAIN_SCREEN;
                  }
                  else if (i == 11)
                  {
                    if (entered >= sensors[selected_sensor].min && 
                        entered <= sensors[selected_sensor].max &&
                        entered <= sensors[selected_sensor].trig_high)
                    {
                      sensors[selected_sensor].trig_low = entered;
                      tft_fillCircle(311, (param + 1)*y_offset + 6, 8, ILI9340_BLACK); 
                      param = (param + 1) % 4;
                      tft_fillCircle(311, (param + 1)*y_offset + 6, 8, ILI9340_GREEN);
                      changed=0;
                    }
                    else
                    {
                      tft_fillCircle(311, (param + 1)*y_offset + 6, 8, ILI9340_RED);
                    
                    }
                    entered = 0;
                  }
                  else if (i >= 0 && i <= 9)
                  {
                     if(!changed){
                        entered=0;
                        changed=1;
                    }  
                    // Display number that has been pressed on the keypad
                    entered = entered*10 + i;
                    tft_fillRoundRect(88, 4*y_offset-5, 200, 25, 1, ILI9340_BLACK);
                    tft_setCursor(96, 4*y_offset);
                    tft_setTextColor(ILI9340_WHITE);
                    sprintf(buffer,"%.2f", entered);
                    tft_writeString(buffer);
                  } 
                  break;
              }
              break;

            case MUSIC_SCREEN:
              switch (i) 
              {
                case 10:
                  draw = STATIC;
                  react_music = 0;
                  screen_fsm = MAIN_SCREEN;
                  colorWipeF(0,0,255); 
                  clearPixels();
                  blue();
                  display();
                  break;

                case 1:
                  // send the prompt via DMA to serial
                  sprintf(PT_send_buffer, "AT-\r\n");
                  PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                  break;

                case 2:
                  // send the prompt via DMA to serial
                  sprintf(PT_send_buffer, "AP\r\n");
                  PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                  break;

                case 3:
                  // send the prompt via DMA to serial
                  sprintf(PT_send_buffer, "AT+\r\n");
                  PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                  break;

                case 4:
                  // send the prompt via DMA to serial
                  sprintf(PT_send_buffer, "AV-\r\n");
                  PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                  break;

                case 5:
                  // send the prompt via DMA to serial
                  sprintf(PT_send_buffer, "AV+\r\n");
                  PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                  break;
              }
              break;

            case GRAPH_SCREEN:
              switch (i) 
              {
                case 10:
                  draw = STATIC;
                  screen_fsm = MAIN_SCREEN;
                  colorWipeF(0,0,255); 
                  display();
                  break;
              }
              break;
          }
          break;
        }
        break;

      case PUSHED:
        state = (press ? PUSHED : MAYBE_NO_PUSH);
        break;

      case MAYBE_NO_PUSH:
        state = (press ? PUSHED : NO_PUSH);
        break;
    }
    
    PT_YIELD_TIME_msec(10);
  }
  PT_END(pt);
}

static PT_THREAD(protothread_led(struct pt *pt)) 
{
  PT_BEGIN(pt);
  static int adc_read;
  static int color = 0;
  static int counter = 10;
  static int color_counter = 0;
  static int i = 0;
  static int q = 0;
  static int r = 0;
  static int b = 0;
  static int g = 0;
  static int forward = 1;
  static int triggered_once = 0;
  while (1) 
  {
    if (react_music)
    {
      if (triggered_once != 1)
      {
        while (buf_power <= 1700)
        {
          red();
          display();
          PT_YIELD_TIME_msec(5);
          if (react_music != 1)
          {
            break;
          }
        }
        triggered_once = 1;
      }
      
      if (react_music)
      {
        if (buf_power > 1700 && counter == 0) 
        {
          forward = (forward + 1) % 2;
          counter = 10;
          color = (color + 1) % 6;
          color_counter = (color_counter + 1) % 6;
          if (color_counter <= 1) 
          {
            r = (char) 80;
            g = (char) 80;
            b = (char) (255 * buf_power / 6000);
          } 
          else if (color_counter <= 3) 
          {
            r = (char) (255 * buf_power / 6000);
            g = (char) 80;
            b = (char) 80;
          } 
          else 
          {
            r = (char) 80;
            g = (char) (255 * buf_power / 6000);
            b = (char) 80;
          }
        }

        if (counter > 0) 
        {
          counter--;
        }

        if (forward) 
        {
          setPixel(i,r,g,b);
          display();
          i = i + 1;
          if (i == 16) 
          {
            i = 0;
          }
        } 
        else 
        {
          setPixel(i,r,g,b);
          display();
          i = i - 1;
          if (i == -1) 
          {
            i = 15;
          }
        }
      }

    }
    else
    {
      triggered_once = 0;
    }
    PT_YIELD_TIME_msec(15);
  }
  PT_END(pt);
}

static PT_THREAD(protothread_measure(struct pt *pt)) 
{
  PT_BEGIN(pt);
  int temp = 0;
  while (1) 
  {
    AcquireADC10();
    
    // Scale down the microphone readout between ~0 and ~200
    temp = (int) (200 * ReadADC10(0) / 1023.0) - 78;
    buffer_pow[buf_idx] = temp;
    buf_idx = (buf_idx + 1) % buf_size;
    int i;
    
    // Calculate average mic power
    for (i = 0; i < buf_size; i++) {
      buf_power += buffer_pow[i] * buffer_pow[i];
    }
    
    // Store result in global variable
    buf_power = buf_power / buf_size;
    
    // Update Luminosity
    get_luminosity(luminosity_values);
    sensors[0].value = (float)luminosity_values[0];
    PT_YIELD_TIME_msec(1);
     
    // Update Temperature and Humidity
    measure_temp_humidity(temp_humidity);
    sensors[1].value = (float)temp_humidity[0];
    sensors[2].value = (float)temp_humidity[1];
    PT_YIELD_TIME_msec(1);
     
    // Update Mic Power
    sensors[3].value = (float)buf_power;
    
    // Update IMU Values
    read_imu_values(imu_values);
    sensors[4].value = (float)imu_values[0];
    sensors[5].value = (float)imu_values[1];
    sensors[6].value = (float)imu_values[2];
    PT_YIELD_TIME_msec(20);
  }
  PT_END(pt);
}

// === I2C Init ================

void i2c_setup(void) 
{
  // SDA1 = Pin 18 RB9
  // SCL1 = Pin 17 RB8
  OpenI2C1(I2C_ON, 0xC2);
}

// Initialize the ADC for reading out potentiometer values
void adc_setup(void) 
{
  // === Potentiometer ADC Setup ===============================================
  // Ensure the ADC is off before setting the configuration
  CloseADC10();

  // ADC_FORMAT_INTG16       Use this in order to get unsigned 16 bit integers
  //
  // ADC_CLK_AUTO            This allows for converting the samples immediately 
  //                         after they are read in
  //
  // ADC_AUTO_SAMPLING_OFF   By setting this, sampling is done when the user 
  //                         wants rather than once the previous sample is done 
  //                         converting
  //
#define PARAM1  ADC_FORMAT_INTG16 | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_OFF

  // ADC_VREF_AVDD_AVSS      Set the voltage references to the same as that of 
  //                         the PIC32 rather than other external references
  //
  // ADC_OFFSET_CAL_DISABLE  We don't want to do the extra steps in order to 
  //                         configure perfect precision as the movement of the 
  //                         paddle is highly relative
  //
  // ADC_SCAN_OFF            Scan mode is disabled because we don't have a 
  //                         vector of inputs to scan through but rather just 
  //                         the one
  //
  // ADC_SAMPLES_PER_INT_1   Only take one sample per interrupt rather than 
  //                         more. We only need to know how the paddle moves 
  //                         once per event
  //
  // ADC_ALT_BUF_OFF         Buffer configured as one 16-word buffer rather 
  //                         than two 8-word buffers
  //
  // ADC_ALT_INPUT_OFF       Always use MUX A input multiplexer settings rather 
  //                         than switching between MUX A and MUX B
  //
#define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_OFF | \
                  ADC_SAMPLES_PER_INT_1 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF

  // ADC_CONV_CLK_PB         Use the peripheral bus clock
  //
  // ADC_SAMPLE_TIME_5       Used to properly set acquisition time
  //
  // ADC_CONV_CLK_Tcy2       Divide the peripheral bus clock by two in order to 
  //                         run it at the highest rate using its 8-bit counter
  //
#define PARAM3 ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_5 | ADC_CONV_CLK_Tcy2 

  // ENABLE_AN11_ANA - enables pin 24 to be our input for the ADC
#define PARAM4	ENABLE_AN11_ANA 

  // SKIP_SCAN_ALL - don't set any channels to scan
#define PARAM5	SKIP_SCAN_ALL

  // ADC_CH0_NEG_SAMPLEA_NVREF  Sets the negative input of the MUX to the 
  //                            negative reference of the PIC32
  //
  // ADC_CH0_POS_SAMPLEA_AN11   Sets the positive input select to be AN11
  //
  SetChanADC10(ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN11);

  // Configure ADC using the parameters defined above
  OpenADC10(PARAM1, PARAM2, PARAM3, PARAM4, PARAM5);

  // Enable the ADC 
  EnableADC10();
}

// Initialize the TFT display and the protothreads threads

void tft_setup(void) 
{
  // Initialize the display
  tft_init_hw();
  tft_begin();

  // Use tft_setRotation(1) for 320x240 resolution
  tft_setRotation(1);

  // UI instantiation
  tft_fillScreen(ILI9340_BLACK);
}

// === Main  ===================================================================

void main(void) 
{
  SYSTEMConfigPerformance(PBCLK);

  // === PT Setup ==============================================================
  PT_setup();

  // === Enable System-Wide Interrupts =========================================
  INTEnableSystemMultiVectoredInt();

  // === ADC Setup =============================================================
  adc_setup();

  // === Initialize the Threads ================================================
  PT_INIT(&pt_input);
  PT_INIT(&pt_timer);
  PT_INIT(&write_neo);
  PT_INIT(&measure_thread);

  // === TFT Setup =============================================================
  tft_setup();
  init_pixels();

  // === I2C Setup =============================================================
  i2c_setup();
  
  // === Sensor Setup ==========================================================
  sensors_setup();
  init_gyro();
  init_luminosity();
  calibrate();
  
  blue();
  display();

  while (1) 
  {
    PT_SCHEDULE(protothread_input(&pt_input));
    PT_SCHEDULE(protothread_timer(&pt_timer));
    PT_SCHEDULE(protothread_led(&write_neo));
    PT_SCHEDULE(protothread_measure(&measure_thread));
  }
}

// === END =====================================================================
